/*
 * Copyright (c) 2018 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License")
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#include <glib.h>
#include <ua-api.h>

#include <user-awareness.h>
#include <user-awareness-log.h>
#include <user-awareness-private.h>
#include <user-awareness-util.h>

GSList *ua_services_list;

static ua_service_info_s* __ua_get_service_from_list(const char *service_name)
{
	GSList *l;
	ua_service_info_s *service_info;

	retv_if(NULL == service_name, NULL);

	for (l = ua_services_list; l; l = g_slist_next(l)) {
		service_info = (ua_service_info_s *)l->data;

		if (!g_strcmp0(service_name, service_info->name))
			return service_info;
	}

	return NULL;
}

void _ua_free_ua_service_info_s(gpointer data)
{
	FUNC_ENTRY;
	ua_service_info_s *service_info = data;

	ret_if(NULL == service_info);

	if (service_info->name)
		g_free(service_info->name);

	g_free(service_info);

	FUNC_EXIT;
}

int _ua_intr_foreach_registered_services(
	_ua_intr_registered_service_cb foreach_cb, void *user_data)
{
	int ret;
	UA_VALIDATE_INPUT_PARAMETER(foreach_cb);

	ret = _ua_foreach_registered_services(foreach_cb, user_data);
	return ret;
}

int _ua_intr_get_default_service(void)
{
	FUNC_ENTRY;
	int ret;
	uam_service_info_s uam_service;
	ua_service_info_s *service = NULL;
	GSList *l;

	ret = _ua_get_error_code(_uam_get_default_service(&uam_service));
	if (UA_ERROR_NONE != ret) {
		/* LCOV_EXCL_START */
		UA_ERR("Failed with error: %s(0x%X)",
				_ua_get_error_string(ret), ret);
		/* LCOV_EXCL_STOP */
		return ret;
	}

	for (l = ua_services_list; l; l = g_slist_next(l)) {
		service = (ua_service_info_s *)l->data;

		if (!g_strcmp0(uam_service.name, service->name)) {
			UA_INFO("Service found [%s]", service->name);
			service->default_service = true;
			service->isadded = true;
			return UA_ERROR_NONE;
		}
	}

	/* LCOV_EXCL_START */
	service = g_malloc0(sizeof(ua_service_info_s));
	if (!service) {
		UA_ERR("Failed to allocate memory");
		return UA_ERROR_OUT_OF_MEMORY;
	}
	service->name = g_strdup(uam_service.name);
	service->service_handle = (ua_service_h)service;
	service->default_service = true;
	service->isadded = true;

	ua_services_list = g_slist_append(ua_services_list, service);
	FUNC_EXIT;
	/* LCOV_EXCL_STOP */
	return UA_ERROR_NONE;
}

GSList *_ua_service_get_services(void)
{
	FUNC_ENTRY;
	FUNC_EXIT;
	return ua_services_list;
}

int _ua_service_add_info_to_list(ua_service_info_s* ua_info)
{
	FUNC_ENTRY;
	ua_service_info_s* service_info;
	UA_VALIDATE_INPUT_PARAMETER(ua_info);

	if (NULL != __ua_get_service_from_list(ua_info->name))
			return UA_ERROR_ALREADY_DONE; // LCOV_EXCL_LINE

	service_info = g_malloc0(sizeof(ua_service_info_s));

	if (!service_info) {
		/* LCOV_EXCL_START */
		UA_ERR("g_malloc0 failed");
		return UA_ERROR_OUT_OF_MEMORY;
		/* LCOV_EXCL_STOP */
	}

	service_info->name = g_strdup(ua_info->name);
	if (service_info->name == NULL) {
		/* LCOV_EXCL_START */
		UA_ERR("g_malloc0 failed");
		_ua_free_ua_service_info_s(service_info);
		return UA_ERROR_OUT_OF_MEMORY;
		/* LCOV_EXCL_STOP */
	}

	service_info->presence_threshold = ua_info->presence_threshold;
	service_info->absence_threshold = ua_info->absence_threshold;
	service_info->isadded = true;
	service_info->service_handle = (ua_service_h)service_info;
	ua_services_list = g_slist_append(ua_services_list, service_info);

	FUNC_EXIT;
	return UA_ERROR_NONE;
}

int _ua_service_add_info_to_list_from_uam_data(uam_service_info_s *uam_info)
{
	FUNC_ENTRY;
	ua_service_info_s* service_info;
	UA_VALIDATE_INPUT_PARAMETER(uam_info);

	if (NULL != __ua_get_service_from_list(uam_info->name)) {
		UA_INFO("Service already in list(added on service_create) [%s]", uam_info->name);
		return UA_ERROR_NONE;
	}

	service_info = g_malloc0(sizeof(ua_service_info_s));
	if (!service_info) {
		UA_ERR("g_malloc0 failed");
		return UA_ERROR_OUT_OF_MEMORY;
	}

	service_info->name = g_strdup(uam_info->name);
	if (service_info->name == NULL) {
		UA_ERR("g_malloc0 failed");
		_ua_free_ua_service_info_s(service_info);
		return UA_ERROR_OUT_OF_MEMORY;
	}

	service_info->isadded = true;
	service_info->service_handle = (ua_service_h)service_info;
	ua_services_list = g_slist_append(ua_services_list, service_info);

	FUNC_EXIT;
	return UA_ERROR_NONE;
}

int _ua_service_remove_info_from_list(uam_service_info_s *uam_info)
{
	FUNC_ENTRY;
	int ret = UA_ERROR_NO_DATA;
	GSList *l;
	ua_service_info_s *s;
	UA_VALIDATE_INPUT_PARAMETER(uam_info);

	for (l = ua_services_list; l; l = g_slist_next(l)) {
		s = (ua_service_info_s *)l->data;

		if (!g_strcmp0(uam_info->name, s->name)) {
			if (s->create_by_app == false) {
				UA_INFO("Service found [%s]", s->name);

				ua_services_list = g_slist_remove(ua_services_list, s);
				_ua_free_ua_service_info_s(s);
			} else {
				UA_INFO("Service created by APP");
				s->isadded = false;
			}
			ret = UA_ERROR_NONE;
			break;
		}
	}

	FUNC_EXIT;
	return ret;
}

int _ua_foreach_registered_services(ua_service_added_service_cb foreach_cb,
	void *user_data)
{
	FUNC_ENTRY;
	int i;
	int ret;
	GPtrArray *service_list = NULL;
	uam_service_info_s *ptr;

	UA_VALIDATE_INPUT_PARAMETER(foreach_cb);

	service_list = g_ptr_array_new();
	retv_if(NULL == service_list, UA_ERROR_OUT_OF_MEMORY);

	ret = _ua_get_error_code(_uam_get_registered_services(&service_list));
	if (UA_ERROR_NONE != ret) {
		/* LCOV_EXCL_START */
		UA_ERR("Failed with error: %s(0x%X)",
				_ua_get_error_string(ret), ret);
		g_ptr_array_free(service_list, TRUE);
		/* LCOV_EXCL_STOP */
		return ret;
	}

	for (i = 0; i < service_list->len; i++) {
		ptr = g_ptr_array_index(service_list, i);
		if (ptr) {
			ua_service_info_s *service_info;
			service_info = g_malloc0(sizeof(ua_service_info_s));

			if (!service_info) {
				/* LCOV_EXCL_START */
				UA_ERR("g_malloc0 failed");
				ret = UA_ERROR_OUT_OF_MEMORY;
				_ua_free_ua_service_info_s(service_info);
				/* LCOV_EXCL_STOP */
				goto done;
			}

			service_info->name = g_strndup(ptr->name, UAM_SERVICE_MAX_STRING_LEN);
			service_info->presence_threshold = ptr->presence_threshold;
			UA_INFO("presence threshold: [%d]", service_info->presence_threshold);
			service_info->absence_threshold = ptr->absence_threshold;
			if (service_info->name == NULL) {
				/* LCOV_EXCL_START */
				UA_ERR("g_malloc0 failed");
				ret = UA_ERROR_OUT_OF_MEMORY;
				_ua_free_ua_service_info_s(service_info);
				/* LCOV_EXCL_STOP */
				goto done;
			}

			service_info->isadded = true;
			service_info->service_handle = (ua_service_h)service_info;
			if (!foreach_cb(service_info->service_handle, user_data)) {
				_ua_free_ua_service_info_s(service_info);
				break;
			} else
				_ua_free_ua_service_info_s(service_info);
		} else {
			/* LCOV_EXCL_START */
			UA_ERR("OPERATION_FAILED(0x%08x)",
					UA_ERROR_OPERATION_FAILED);
			ret = UA_ERROR_OPERATION_FAILED;
			/* LCOV_EXCL_STOP */
			goto done;
		}
	}

done:
	g_ptr_array_foreach(service_list, (GFunc)g_free, NULL);
	g_ptr_array_free(service_list, TRUE);

	FUNC_EXIT;
	return ret;
}


int ua_service_create(ua_service_h *service_handle)
{
	FUNC_ENTRY;
	ua_service_info_s *service = NULL;

	UA_VALIDATE_INPUT_PARAMETER(service_handle);

	service = g_malloc0(sizeof(ua_service_info_s));
	if (!service) {
		/* LCOV_EXCL_START */
		UA_ERR("g_malloc0 failed");
		return UA_ERROR_OUT_OF_MEMORY;
		/* LCOV_EXCL_STOP */
	}
	service->name = NULL;

	/* Add service to list of services*/
	*service_handle = (ua_service_h)service;
	service->service_handle = (ua_user_h)service;
	service->create_by_app = true;

	ua_services_list = g_slist_append(ua_services_list, service);

	FUNC_EXIT;
	return UA_ERROR_NONE;
}

int ua_service_destroy(ua_service_h service_handle)
{
	FUNC_ENTRY;
	ua_service_info_s *service = (ua_service_info_s *)service_handle;

	UA_VALIDATE_INPUT_PARAMETER(service_handle);
	UA_VALIDATE_HANDLE(service_handle, ua_services_list);
	retv_if(FALSE == service->create_by_app, UA_ERROR_INVALID_PARAMETER);

	/* Do not remove from list if service is added to database */
	if (service->isadded) {
		/* LCOV_EXCL_START */
		service->create_by_app = FALSE;
		return UA_ERROR_NONE;
		/* LCOV_EXCL_STOP */
	}

	ua_services_list = g_slist_remove(ua_services_list, service);
	_ua_free_ua_service_info_s(service);

	FUNC_EXIT;
	return UA_ERROR_NONE;
}

int ua_service_add(ua_service_h service_handle)
{
	FUNC_ENTRY;

	int ret;
	uam_service_info_s uam_service;
	ua_service_info_s *service = (ua_service_info_s *)service_handle;

	UA_VALIDATE_INPUT_PARAMETER(service_handle);
	UA_VALIDATE_HANDLE(service_handle, ua_services_list);

	retv_if(NULL == service->name, UA_ERROR_INVALID_PARAMETER);
	retv_if(service->isadded == TRUE, UA_ERROR_ALREADY_DONE);

	memset(&uam_service, 0, sizeof(uam_service_info_s));
	g_strlcpy(uam_service.name, service->name, UAM_SERVICE_MAX_STRING_LEN);
	uam_service.presence_threshold = service->presence_threshold;
	uam_service.absence_threshold = service->absence_threshold;

	ret = _ua_get_error_code(_uam_register_service(&uam_service));
	if (UA_ERROR_NONE != ret) {
		/* LCOV_EXCL_START */
		UA_ERR("Failed with error: %s(0x%X)",
				_ua_get_error_string(ret), ret);
		return ret;
		/* LCOV_EXCL_STOP */
	}

	service->isadded = true;
	FUNC_EXIT;
	return UA_ERROR_NONE;
}

int ua_service_update(ua_service_h service_handle)
{
	FUNC_ENTRY;

	int ret;
	uam_service_info_s uam_service;
	ua_service_info_s *service = (ua_service_info_s *)service_handle;

	UA_VALIDATE_INPUT_PARAMETER(service_handle);
	UA_VALIDATE_HANDLE(service_handle, ua_services_list);

	retv_if(NULL == service->name, UA_ERROR_INVALID_PARAMETER);
	retv_if(service->isadded != TRUE, UA_ERROR_INVALID_PARAMETER);

	memset(&uam_service, 0, sizeof(uam_service_info_s));
	g_strlcpy(uam_service.name, service->name, UAM_SERVICE_MAX_STRING_LEN);
	uam_service.presence_threshold = service->presence_threshold;
	uam_service.absence_threshold = service->absence_threshold;

	ret = _ua_get_error_code(_uam_update_service(&uam_service));
	if (UA_ERROR_NONE != ret) {
		/* LCOV_EXCL_START */
		UA_ERR("Failed with error: %s(0x%X)",
				_ua_get_error_string(ret), ret);
		return ret;
		/* LCOV_EXCL_STOP */
	}

	FUNC_EXIT;
	return UA_ERROR_NONE;
}

int ua_service_remove(ua_service_h service_handle)
{
	FUNC_ENTRY;
	int ret;
	uam_service_info_s uam_service;
	ua_service_info_s *service = (ua_service_info_s *)service_handle;

	UA_VALIDATE_INPUT_PARAMETER(service_handle);
	UA_VALIDATE_HANDLE(service_handle, ua_services_list);
	retv_if(NULL == service->name, UA_ERROR_INVALID_PARAMETER);
	retv_if(service->isadded == FALSE, UA_ERROR_INVALID_PARAMETER);

	memset(&uam_service, 0, sizeof(uam_service_info_s));
	g_strlcpy(uam_service.name, service->name, UAM_SERVICE_MAX_STRING_LEN);

	ret = _ua_get_error_code(_uam_unregister_service(&uam_service));
	if (UA_ERROR_NONE != ret) {
		/* LCOV_EXCL_START */
		UA_ERR("Failed with error: %s(0x%X)",
				_ua_get_error_string(ret), ret);
		return ret;
		/* LCOV_EXCL_STOP */
	}

	service->isadded = false;

	FUNC_EXIT;
	return UA_ERROR_NONE;
}

int ua_service_set_name(ua_service_h service_handle, const char *name)
{
	FUNC_ENTRY;
	ua_service_info_s *service = (ua_service_info_s *)service_handle;

	UA_VALIDATE_INPUT_PARAMETER(name);
	UA_VALIDATE_INPUT_PARAMETER(service_handle);
	UA_VALIDATE_HANDLE(service_handle, ua_services_list);

	retv_if(service->isadded, UA_ERROR_NOT_PERMITTED);
	retv_if((strlen(name) > UAM_SERVICE_MAX_STRING_LEN), UA_ERROR_INVALID_PARAMETER);

	if (service->name)
		g_free(service->name); // LCOV_EXCL_LINE
	service->name = g_strdup(name);

	FUNC_EXIT;
	return UA_ERROR_NONE;
}

int ua_service_set_detection_threshold(ua_service_h service_handle,
	unsigned int presence_threshold, unsigned int absence_threshold)
{
	FUNC_ENTRY;
	ua_service_info_s *service = (ua_service_info_s *)service_handle;

	UA_VALIDATE_INPUT_PARAMETER(service_handle);
	UA_VALIDATE_HANDLE(service_handle, ua_services_list);

	service->presence_threshold = presence_threshold;
	service->absence_threshold = absence_threshold;

	FUNC_EXIT;
	return UA_ERROR_NONE;
}

int ua_service_get_name(ua_service_h service_handle, char **name)
{
	FUNC_ENTRY;
	ua_service_info_s *service = (ua_service_info_s *)service_handle;

	UA_VALIDATE_INPUT_PARAMETER(name);
	UA_VALIDATE_INPUT_PARAMETER(service_handle);
	UA_VALIDATE_HANDLE(service_handle, ua_services_list);

	retv_if(NULL == service->name, UA_ERROR_NO_DATA);

	*name = g_strdup(service->name);
	if (*name == NULL) {
		/* LCOV_EXCL_START */
		UA_ERR("g_malloc0 failed");
		return UA_ERROR_OUT_OF_MEMORY;
		/* LCOV_EXCL_STOP */
	}

	FUNC_EXIT;
	return UA_ERROR_NONE;
}

int ua_service_get_detection_threshold(ua_service_h service_handle,
	unsigned int *presence_threshold, unsigned int *absence_threshold)
{
	FUNC_ENTRY;
	ua_service_info_s *service = (ua_service_info_s *)service_handle;

	UA_VALIDATE_INPUT_PARAMETER(service_handle);
	UA_VALIDATE_HANDLE(service_handle, ua_services_list);

	retv_if(NULL == service->name, UA_ERROR_NO_DATA);

	*presence_threshold = service->presence_threshold;
	*absence_threshold = service->absence_threshold;

	FUNC_EXIT;
	return UA_ERROR_NONE;
}

int ua_service_get_default_service(ua_service_h* service_handle)
{
	FUNC_ENTRY;
	int ret;
	uam_service_info_s uam_service;
	ua_service_info_s *h_service = NULL;
	GSList *l;

	UA_VALIDATE_INPUT_PARAMETER(service_handle);

	for (l = ua_services_list; l; l = g_slist_next(l)) {
		h_service = (ua_service_info_s *)l->data;

		if (h_service->default_service == true) {
			/* LCOV_EXCL_START */
			UA_INFO("Service found [%s]", h_service->name);
			*service_handle = (ua_user_h)h_service;
			return UA_ERROR_NONE;
			/* LCOV_EXCL_STOP */
		}
	}

	ret = _ua_get_error_code(_uam_get_default_service(&uam_service));
	if (UA_ERROR_NONE != ret) {
		/* LCOV_EXCL_START */
		UA_ERR("Failed with error: %s(0x%X)",
				_ua_get_error_string(ret), ret);
		*service_handle = NULL;
		return ret;
		/* LCOV_EXCL_STOP */
	}

	h_service = g_malloc0(sizeof(ua_service_info_s));
	if (!h_service) {
		/* LCOV_EXCL_START */
		UA_ERR("Failed to allocated memory");
		return UA_ERROR_OUT_OF_MEMORY;
		/* LCOV_EXCL_STOP */
	}
	h_service->name = g_strdup(uam_service.name);
	h_service->service_handle = (ua_user_h)h_service;

	*service_handle = (ua_service_h)h_service;
	h_service->default_service = true;
	h_service->isadded = true;
	ua_services_list = g_slist_append(ua_services_list, h_service);
	FUNC_EXIT;
	return UA_ERROR_NONE;
}

int ua_service_get_by_name(const char *name, ua_service_h *service_handle)
{
	FUNC_ENTRY;
	GSList *l;
	ua_service_info_s *s;

	UA_VALIDATE_INPUT_PARAMETER(name);
	UA_VALIDATE_INPUT_PARAMETER(service_handle);

	for (l = ua_services_list; l; l = g_slist_next(l)) {
		s = (ua_service_info_s *)l->data;

		if (!g_strcmp0(name, s->name)) {
			UA_INFO("Service found [%s]", s->name);
			*service_handle = s;
			break;
		}
	}

	if(l == NULL) {
			UA_ERR("Service [%s] is not found", name);
			FUNC_EXIT;
			return UA_ERROR_INVALID_PARAMETER;
	}

	FUNC_EXIT;
	return UA_ERROR_NONE;
}

int ua_service_add_user(ua_service_h service_handle, ua_user_h user_handle)
{
	FUNC_ENTRY;
	int ret;

	ua_service_info_s* service_info = (ua_service_info_s*)service_handle;
	ua_user_info_s* user_info = (ua_user_info_s*)user_handle;
	GSList *user_list = _ua_user_get_users();
	ua_service_info_s *tmp = NULL;

	UA_VALIDATE_INPUT_PARAMETER(service_handle);
	UA_VALIDATE_INPUT_PARAMETER(user_handle);
	UA_VALIDATE_HANDLE(user_handle, user_list);
	UA_VALIDATE_HANDLE(service_handle, ua_services_list);
	retv_if(user_info->isadded == FALSE, UA_ERROR_INVALID_PARAMETER);

#if 0
//TODO lk, ideally the handle should be in the ua_services_list.
	tmp = __ua_get_service_from_list(service_info->name);
	if (!tmp) {
		ua_service_info_s *tmp = g_malloc0(sizeof(ua_service_info_s));
		if (!tmp) {
			UA_ERR("Memory allocation failed");
			return UA_ERROR_OUT_OF_MEMORY;
		}
		tmp->name = g_strdup(service_info->name);
		tmp->create_by_app = true;
		tmp->default_service = false;
		tmp->isadded = true;
		tmp->service_handle = tmp;
		ua_services_list = g_slist_append(ua_services_list, tmp);
	}
#else
	tmp = service_info;
#endif

	ret = _ua_get_error_code(_uam_service_add_user(tmp->name, user_info->account));
	if (UA_ERROR_NONE != ret) {
		UA_ERR("_uam_service_add_user returned %s",
				_ua_get_error_string(ret));
		return ret;
	}

	FUNC_EXIT;
	return UA_ERROR_NONE;
}

int ua_service_remove_user(ua_service_h service_handle, ua_user_h user_handle)
{
	FUNC_ENTRY;
	int ret;

	ua_service_info_s* service_info = (ua_service_info_s*)service_handle;
	ua_user_info_s* user_info = (ua_user_info_s*)user_handle;
	GSList *user_list = _ua_user_get_users();

	UA_VALIDATE_INPUT_PARAMETER(service_info);
	UA_VALIDATE_INPUT_PARAMETER(user_handle);
	UA_VALIDATE_HANDLE(service_handle, ua_services_list);
	UA_VALIDATE_HANDLE(user_handle, user_list);

	ret = _ua_get_error_code(_uam_service_remove_user(service_info->name, user_info->account));
	if (UA_ERROR_NONE != ret) {
		UA_ERR("_uam_service_remove_user returned %s",
				_ua_get_error_string(ret));
		return ret;
	}

	FUNC_EXIT;
	return UA_ERROR_NONE;
}

int ua_service_add_device(ua_service_h service_handle, ua_device_h device_handle)
{
	FUNC_ENTRY;
	int ret;

	ua_service_info_s* service_info = (ua_service_info_s*)service_handle;
	ua_dev_info_s* device_info = (ua_dev_info_s*)device_handle;
	ua_service_info_s *tmp = NULL;

	UA_VALIDATE_INPUT_PARAMETER(service_info);
	UA_VALIDATE_INPUT_PARAMETER(device_handle);
	UA_VALIDATE_HANDLE(service_handle, ua_services_list);
	UA_PRINT_DEVICE_HANDLE(device_handle);
	retv_if(device_info->isadded == FALSE, UA_ERROR_INVALID_PARAMETER);

#if 0
//TODO lk, ideally the handle should be in the ua_services_list.
	tmp = __ua_get_service_from_list(service_info->name);
	if (!tmp) {
		ua_service_info_s *tmp = g_malloc0(sizeof(ua_service_info_s));
		if (!tmp) {
			UA_ERR("Memory allocation failed");
			return UA_ERROR_OUT_OF_MEMORY;
		}
		tmp->name = g_strdup(service_info->name);
		tmp->create_by_app = true;
		tmp->default_service = false;
		tmp->isadded = true;
		tmp->service_handle = tmp;
		ua_services_list = g_slist_append(ua_services_list, tmp);
	}
#else
	tmp = service_info;
#endif
	ret = _ua_get_error_code(_uam_service_add_device(tmp->name, device_info->device_id,
									_ua_to_uam_tech_type(device_info->type)));
	if (UA_ERROR_NONE != ret) {
		UA_ERR("_uam_service_add_device returned %s",
				_ua_get_error_string(ret));
		return ret;
	}

	FUNC_EXIT;
	return UA_ERROR_NONE;
}

int ua_service_remove_device(ua_service_h service_handle, ua_device_h device_handle)
{
	FUNC_ENTRY;
	int ret;

	ua_service_info_s* service_info = (ua_service_info_s*)service_handle;
	ua_dev_info_s* device_info = (ua_dev_info_s*)device_handle;

	UA_VALIDATE_INPUT_PARAMETER(service_info);
	UA_VALIDATE_INPUT_PARAMETER(service_info->name);
	UA_VALIDATE_INPUT_PARAMETER(device_info);
	UA_VALIDATE_INPUT_PARAMETER(device_info->device_id);

	UA_VALIDATE_HANDLE(service_handle, ua_services_list);
	UA_PRINT_DEVICE_HANDLE(device_handle);

	ret = _ua_get_error_code(_uam_service_remove_device(service_info->name,
			device_info->device_id, _ua_to_uam_tech_type(device_info->type)));
	if (UA_ERROR_NONE != ret) {
		UA_ERR("_uam_service_remove_device returned %s",
				_ua_get_error_string(ret));
		return ret;
	}

	FUNC_EXIT;
	return UA_ERROR_NONE;
}

int ua_service_set_device_discriminant(ua_service_h service_handle,
	ua_device_h device_handle, gboolean discriminant)
{
	FUNC_ENTRY;
	int ret;

	ua_service_info_s* service_info = (ua_service_info_s*)service_handle;
	ua_dev_info_s* device_info = (ua_dev_info_s*)device_handle;

	UA_VALIDATE_INPUT_PARAMETER(service_info);
	UA_VALIDATE_INPUT_PARAMETER(device_handle);
	UA_VALIDATE_HANDLE(service_handle, ua_services_list);
	UA_PRINT_DEVICE_HANDLE(device_handle);
	retv_if(device_info->isadded == FALSE, UA_ERROR_INVALID_PARAMETER);

	ret = _ua_get_error_code(_uam_service_set_device_discriminant(service_info->name,
				device_info->device_id, _ua_to_uam_tech_type(device_info->type),
				discriminant));
	if (UA_ERROR_NONE != ret) {
		UA_ERR("_uam_service_set_device_discriminant returned %s",
				_ua_get_error_string(ret));
		return ret;
	}

	FUNC_EXIT;
	return UA_ERROR_NONE;
}

int ua_service_get_device_discriminant(ua_service_h service_handle,
	ua_device_h device_handle, gboolean *discriminant)
{
	FUNC_ENTRY;
	int ret;

	ua_service_info_s* service_info = (ua_service_info_s*)service_handle;
	ua_dev_info_s* device_info = (ua_dev_info_s*)device_handle;

	UA_VALIDATE_INPUT_PARAMETER(service_info);
	UA_VALIDATE_INPUT_PARAMETER(device_handle);
	UA_VALIDATE_HANDLE(service_handle, ua_services_list);
	UA_PRINT_DEVICE_HANDLE(device_handle);
	retv_if(device_info->isadded == FALSE, UA_ERROR_INVALID_PARAMETER);

	ret = _ua_get_error_code(_uam_service_get_device_discriminant(service_info->name,
				device_info->device_id, _ua_to_uam_tech_type(device_info->type),
				discriminant));
	if (UA_ERROR_NONE != ret) {
		UA_ERR("_uam_service_get_device_discriminant returned %s",
				_ua_get_error_string(ret));
		return ret;
	}

	FUNC_EXIT;
	return UA_ERROR_NONE;
}

int ua_service_foreach_added_services(ua_service_added_service_cb foreach_cb,
		void *user_data)
{
	FUNC_ENTRY;
	GSList *l;
	ua_service_info_s *s;

	UA_VALIDATE_INPUT_PARAMETER(foreach_cb);

	for (l = ua_services_list; l; l = g_slist_next(l)) {
		s = (ua_service_info_s *)l->data;

//TODO lk, send a copy of handle, which is valid in scope of callback method only
		if (!foreach_cb(s->service_handle, user_data))
			break;
	}

	FUNC_EXIT;
	return UA_ERROR_NONE;
}

int ua_service_foreach_added_users(
		ua_service_h service_handle,
		ua_service_added_user_cb foreach_cb,
		void *user_data)
{
	FUNC_ENTRY;
	int i;
	int ret;
	ua_service_info_s *service = (ua_service_info_s *)service_handle;
	GPtrArray *users_list = NULL;
	uam_user_info_s *ptr;

	UA_VALIDATE_INPUT_PARAMETER(foreach_cb);
	UA_VALIDATE_INPUT_PARAMETER(service_handle);
	UA_VALIDATE_HANDLE(service_handle, ua_services_list);

	users_list = g_ptr_array_new();
	retv_if(NULL == users_list, UA_ERROR_OUT_OF_MEMORY);

	ret = _ua_get_error_code(_uam_request_get_service_users(service->name, &users_list));
	if (UA_ERROR_NONE != ret) {
		UA_ERR("Failed with error: %s(0x%X)",
				_ua_get_error_string(ret), ret);
		g_ptr_array_free(users_list, TRUE);
		return ret;
	}

	for (i = 0; i < users_list->len; i++) {
		/* LCOV_EXCL_START */
		ptr = g_ptr_array_index(users_list, i);
		if (ptr) {
			ua_user_info_s* user_info;
			user_info = g_malloc0(sizeof(ua_user_info_s));
			if (!user_info) {
				UA_ERR("g_malloc0 failed");
				ret = UA_ERROR_OUT_OF_MEMORY;
				goto done;
			}

			user_info->account = g_strdup(ptr->account);
			if (user_info->account == NULL) {
				UA_ERR("g_malloc0 failed");
				ret = UA_ERROR_OUT_OF_MEMORY;
				_ua_free_ua_user_info_t(user_info);
				goto done;
			}

			user_info->name = g_strdup(ptr->name);
			if (user_info->name == NULL) {
				UA_ERR("g_malloc0 failed");
				ret = UA_ERROR_OUT_OF_MEMORY;
				_ua_free_ua_user_info_t(user_info);
				goto done;
			}

			user_info->isadded = true;
			user_info->user_handle = (ua_user_h)user_info;
			user_info->state = UA_PRSENCE_STATE_INVALID;

			if (!foreach_cb(service->service_handle, user_info->user_handle, user_data)) {
				_ua_free_ua_user_info_t(user_info);
				break;
			} else
				_ua_free_ua_user_info_t(user_info);
		} else {
			UA_ERR("OPERATION_FAILED(0x%08x)",
					UA_ERROR_OPERATION_FAILED);
			ret = UA_ERROR_OPERATION_FAILED;
			goto done;
		}
		/* LCOV_EXCL_STOP */
	}

done:
	g_ptr_array_foreach(users_list, (GFunc)g_free, NULL);
	g_ptr_array_free(users_list, TRUE);

	FUNC_EXIT;
	return ret;
}

int ua_service_foreach_added_devices(
		ua_service_h service_handle,
		ua_service_added_device_cb foreach_cb,
		void *user_data)
{
	FUNC_ENTRY;
	int i;
	int ret;
	ua_service_info_s *service = (ua_service_info_s *)service_handle;
	GPtrArray *devices_list = NULL;
	uam_device_info_s *ptr;

	UA_VALIDATE_INPUT_PARAMETER(foreach_cb);
	UA_VALIDATE_INPUT_PARAMETER(service_handle);
	UA_VALIDATE_HANDLE(service_handle, ua_services_list);

	devices_list = g_ptr_array_new();
	retv_if(NULL == devices_list, UA_ERROR_OUT_OF_MEMORY);

	ret = _ua_get_error_code(_uam_request_get_service_devices(service->name, &devices_list));
	if (UA_ERROR_NONE != ret) {
		UA_ERR("Failed with error: %s(0x%X)",
				_ua_get_error_string(ret), ret);
		g_ptr_array_free(devices_list, TRUE);
		return ret;
	}

	for (i = 0; i < devices_list->len; i++) {
		/* LCOV_EXCL_START */
		ptr = g_ptr_array_index(devices_list, i);
		if (ptr) {
			ua_dev_info_s* device_info;
			device_info = g_malloc0(sizeof(ua_dev_info_s));
			if (!device_info) {
				UA_ERR("g_malloc0 failed");
				ret = UA_ERROR_OUT_OF_MEMORY;
				goto done;
			}

			device_info->mac = g_strdup(ptr->mac);
			if (device_info->mac == NULL) {
				UA_ERR("g_malloc0 failed");
				ret = UA_ERROR_OUT_OF_MEMORY;
				_ua_free_ua_device_info_t((gpointer)device_info);
				goto done;
			}

			device_info->ipv4 = g_strdup(ptr->ipv4_addr);
			if (device_info->ipv4 == NULL) {
				UA_ERR("g_malloc0 failed");
				ret = UA_ERROR_OUT_OF_MEMORY;
				_ua_free_ua_device_info_t((gpointer)device_info);
				goto done;
			}

			device_info->device_id = g_strdup(ptr->device_id);
			if (device_info->device_id == NULL) {
				UA_ERR("g_malloc0 failed");
				ret = UA_ERROR_OUT_OF_MEMORY;
				_ua_free_ua_device_info_t((gpointer)device_info);
				goto done;
			}

			device_info->payload = _ua_get_payload_from_uam(&ptr->payload);
			if (device_info->payload == NULL) {
				UA_ERR("_ua_get_payload_from_uam() failed");
				ret = UA_ERROR_OUT_OF_MEMORY;
				_ua_free_ua_device_info_t((gpointer)device_info);
				goto done;
			}

			device_info->isadded = true;
			device_info->handle = (ua_device_h)device_info;
			device_info->type = _to_ua_mac_type(ptr->type);
			device_info->os = ptr->operating_system;
			device_info->discriminant = ptr->discriminant;
			device_info->user = NULL;

			if (!foreach_cb(service->service_handle, device_info->handle, user_data)) {
				_ua_free_ua_device_info_t(device_info);
				break;
			} else
				_ua_free_ua_device_info_t(device_info);

		} else {
			UA_ERR("OPERATION_FAILED(0x%08x)",
					UA_ERROR_OPERATION_FAILED);
			ret = UA_ERROR_OPERATION_FAILED;
			goto done;
		}
		/* LCOV_EXCL_STOP */
	}

done:
	g_ptr_array_foreach(devices_list, (GFunc)g_free, NULL);
	g_ptr_array_free(devices_list, TRUE);

	FUNC_EXIT;
	return ret;
}

