/*
 * Copyright (c) 2018 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <glib.h>
#include <gio/gio.h>

#include <user-awareness.h>

#include "uat-menu.h"
#include "uat-common.h"

ua_service_h g_service_h = NULL; /**< Service handle */
char g_service_str[MENU_DATA_SIZE + 1] = { "ua.service.default" }; /**< Seletected service name */

static char g_presence_threshold_str[MENU_DATA_SIZE + 1] = {0,}; /**< Seletected service presence threshold */
static char g_absence_threshold_str[MENU_DATA_SIZE + 1] = {0,}; /**< Seletected service absence threshold */
static char g_selected_service_id[MENU_DATA_SIZE + 1] = {0,}; /**< Selected service id in the list */

static GSList *g_service_list = NULL; /**< Service List */
static int g_service_count = 0; /**< Service count */

static void __update_service_info(void)
{
	int ret = UA_ERROR_NONE;
	char *name = NULL;

	if (1 == g_service_count) {
		if (g_service_h) {
			ua_service_destroy(g_service_h);
			g_service_h = NULL;
		}

		ua_service_get_default_service(&g_service_h);
	}

	ret = ua_service_get_name(g_service_h, &name);
	if (UA_ERROR_NONE == ret) {
		memset(g_service_str, 0, MENU_DATA_SIZE + 1);
		memcpy(g_service_str, name, MENU_DATA_SIZE);
		g_free(name);
	}
}

static void  ___free_service(gpointer data)
{
	ua_service_h handle = data;
	ua_service_destroy(handle);
	handle = NULL;
}

static void __clear_service_list(void)
{
	g_slist_free_full(g_service_list, ___free_service);
	g_service_list = NULL;
	g_service_count = 0;
}

static bool __foreach_added_service_cb(
	ua_service_h service_handle, void *user_data)
{
	int ret = UA_ERROR_NONE;
	char *name = NULL;
	unsigned int presence_threshold = 0;
	unsigned int absence_threshold = 0;

	check_if(NULL == service_handle);

	ret = ua_service_get_name(service_handle, &name);
	if (UA_ERROR_NONE != ret) {
		msg(" - ua_service_get_name() ret: [0x%X] [%s]",
			ret, uat_get_error_str(ret));
		return RET_SUCCESS;
	}

	ret = ua_service_get_detection_threshold(service_handle,
			&presence_threshold, &absence_threshold);
	if (UA_ERROR_NONE != ret) {
		msg(" - ua_service_get_detection_threshold() ret: [0x%X] [%s]",
			ret, uat_get_error_str(ret));
		g_free(name);
		return RET_SUCCESS;
	}

	msgb("[%d] %s ", ++g_service_count, name);
	msgb("presence threshold :[%d] absence threshold :[%d]",
			presence_threshold, absence_threshold);

	g_service_list = g_slist_append(g_service_list, service_handle);

	g_free(name);

	return true;
}


static int run_ua_service_get_default_service(MManager *mm, struct menu_data *menu)
{
	int ret = UA_ERROR_NONE;

	msg("ua_service_get_default_service");

	if (g_service_h) {
		ua_service_destroy(g_service_h);
		g_service_h = NULL;
	}

	ret = ua_service_get_default_service(&g_service_h);

	if (UA_ERROR_NONE == ret)
		__update_service_info();

	msg(" - ua_service_get_default_service() ret: [0x%X] [%s]",
		ret, uat_get_error_str(ret));

	return RET_SUCCESS;
}

static int run_ua_service_create(
	MManager *mm, struct menu_data *menu)
{
	int ret = UA_ERROR_NONE;

	msg("ua_service_create");

	if (g_service_h) {
		ua_service_destroy(g_service_h);
		g_service_h = NULL;
	}

	ret = ua_service_create(&g_service_h);

	msg(" - ua_service_create() ret: [0x%X] [%s]",
			ret, uat_get_error_str(ret));

	return RET_SUCCESS;
}

static int run_ua_service_destroy(MManager *mm, struct menu_data *menu)
{
	int ret = UA_ERROR_NONE;

	msg("ua_service_destroy");

	ret = ua_service_destroy(g_service_h);
	g_service_h = NULL;

	msg(" - ua_service_destroy() ret: [0x%X] [%s]",
			ret, uat_get_error_str(ret));

	return RET_SUCCESS;
}

static int run_ua_service_set_name(MManager *mm, struct menu_data *menu)
{
	int ret = UA_ERROR_NONE;

	msg("ua_service_set_name");

	check_if(NULL == g_service_h);

	ret = ua_service_set_name(g_service_h, g_service_str);

	msg(" - ua_service_set_name() ret: [0x%X] [%s]",
			ret, uat_get_error_str(ret));

	return RET_SUCCESS;
}

static int run_ua_service_set_detection_threshold(MManager *mm, struct menu_data *menu)
{
	int ret = UA_ERROR_NONE;
	unsigned int presence_threshold = 60;
	unsigned int absence_threshold = 60;

	msg("ua_service_set_threshold");

	check_if(NULL == g_service_h);

	if (strlen(g_presence_threshold_str))
		presence_threshold = (unsigned int)strtol(g_presence_threshold_str, NULL, 10);
	if (strlen(g_absence_threshold_str))
		absence_threshold = (unsigned int)strtol(g_absence_threshold_str, NULL, 10);

	ret = ua_service_set_detection_threshold(g_service_h, presence_threshold,
					absence_threshold);

	msg(" - ua_service_set_name() ret: [0x%X] [%s]",
			ret, uat_get_error_str(ret));

	return RET_SUCCESS;
}

static int run_ua_service_add(
	MManager *mm, struct menu_data *menu)
{
	int ret = UA_ERROR_NONE;

	msg("ua_service_add");

	check_if(NULL == g_service_h);

	ret = ua_service_add(g_service_h);

	msg(" - ua_service_add() ret: [0x%X] [%s]",
			ret, uat_get_error_str(ret));

	return RET_SUCCESS;
}

static int run_ua_service_update(
	MManager *mm, struct menu_data *menu)
{
	int ret = UA_ERROR_NONE;

	msg("ua_service_update");

	check_if(NULL == g_service_h);

	ret = ua_service_update(g_service_h);

	msg(" - ua_service_update() ret: [0x%X] [%s]",
			ret, uat_get_error_str(ret));

	return RET_SUCCESS;
}

static int run_ua_service_remove(
	MManager *mm, struct menu_data *menu)
{
	int ret = UA_ERROR_NONE;

	msg("ua_service_remove");

	check_if(NULL == g_service_h);

	ret = ua_service_remove(g_service_h);

	msg(" - ua_service_remove() ret: [0x%X] [%s]",
			ret, uat_get_error_str(ret));

	return RET_SUCCESS;
}

static int run_ua_service_get_by_name(MManager *mm, struct menu_data *menu)
{
	int ret = UA_ERROR_NONE;

	msg("ua_service_get_by_name");

	if (g_service_h) {
		ua_service_destroy(g_service_h);
		g_service_h = NULL;
	}

	ret = ua_service_get_by_name(g_service_str, &g_service_h);

	if (UA_ERROR_NONE == ret)
		__update_service_info();

	msg(" - ua_service_get_by_name() ret: [0x%X] [%s]",
		ret, uat_get_error_str(ret));

	return RET_SUCCESS;
}

static int run_ua_service_foreach_added_services(
	MManager *mm, struct menu_data *menu)
{
	int ret = UA_ERROR_NONE;
	msg("ua_service_foreach_added_services");

	__clear_service_list();

	ret = ua_service_foreach_added_services(__foreach_added_service_cb, NULL);

	msg(" - ua_service_foreach_added_services() ret: [0x%X] [%s]",
				ret, uat_get_error_str(ret));

	return RET_SUCCESS;
}

static struct menu_data menu_ua_service_set_name[] = {
	{ "1", "name", NULL,
		NULL, g_service_str },
	{ "2", "run", NULL,
		run_ua_service_set_name, NULL },
	{ NULL, NULL, },
};

static struct menu_data menu_ua_service_get_by_name[] = {
	{ "1", "name", NULL,
		NULL, g_service_str },
	{ "2", "run", NULL,
		run_ua_service_get_by_name, NULL },
	{ NULL, NULL, },
};

static struct menu_data menu_ua_service_set_detection_threshold[] = {
	{ "1", "light presence detection threshold",
		NULL, NULL, g_presence_threshold_str},
	{ "2", "light absence detection threshold",
		NULL, NULL, g_absence_threshold_str},
	{ "3", "run",
		NULL, run_ua_service_set_detection_threshold, NULL },
	{ NULL, NULL, },
};

struct menu_data menu_ua_services[] = {
	{ "1", "ua_service_get_default_service",
		NULL, run_ua_service_get_default_service, NULL },
	{ "2", "ua_service_create",
		NULL, run_ua_service_create, NULL },
	{ "3", "ua_service_set_name",
		menu_ua_service_set_name, NULL, g_service_str },
	{ "4", "ua_service_set_detection_threshold",
		menu_ua_service_set_detection_threshold, NULL, NULL },
	{ "5", "ua_service_add",
		NULL, run_ua_service_add, NULL },
	{ "6", "ua_service_update",
		NULL, run_ua_service_update, NULL },
	{ "7", "ua_service_remove",
		NULL, run_ua_service_remove, NULL},
	{ "8", "ua_service_destroy",
		NULL, run_ua_service_destroy, NULL},
	{ "9", "ua_service_get_by_name",
		menu_ua_service_get_by_name, NULL, NULL },
	{ "10", "ua_service_foreach_added_services",
		NULL, run_ua_service_foreach_added_services, NULL },
	{ NULL, NULL, },
};

static int run_choose_service_list(MManager *mm, struct menu_data *menu)
{
	__clear_service_list();
	ua_service_foreach_added_services(__foreach_added_service_cb, NULL);
	return RET_SUCCESS;
}

static int run_select_service(MManager *mm, struct menu_data *menu)
{
	GSList *iter = g_service_list;

	int id = 0;
	int selected_id = 0;

	if (strlen(g_selected_service_id))
		selected_id = (unsigned char)strtol(g_selected_service_id, NULL, 10);

	if (selected_id <= 0) {
		msg("Please select service first");
		return RET_SUCCESS;
	}

	while (NULL != iter) {
		GSList *next = iter->next;
		ua_service_h *handle = iter->data;
		if (handle && ++id == selected_id) {
			/* Make selected service as a current one */
			g_service_h = handle;
			__update_service_info();
			msg("Select service done");
			return RET_SUCCESS;
		}
		iter = next;
	}

	return RET_SUCCESS;
}

struct menu_data menu_sel_service[] = {
	{ "1", "Service list", NULL, run_choose_service_list, g_selected_service_id},
	{ "2", "Apply", NULL, run_select_service, NULL },
	{ NULL, NULL, },
};
